package java.lang.invoke;

import java.io.Serializable;
import java.lang.reflect.Array;
import java.util.Arrays;

class Invokers
{
    private final MethodType targetType;
    @Stable
    private final MethodHandle[] invokers;
    static final int INV_EXACT = 0;
    static final int INV_GENERIC = 1;
    static final int INV_BASIC = 2;
    static final int INV_LIMIT = 3;
    private static final int MH_LINKER_ARG_APPENDED = 1;
    private static final LambdaForm.NamedFunction NF_checkExactType;
    private static final LambdaForm.NamedFunction NF_checkGenericType;
    private static final LambdaForm.NamedFunction NF_getCallSiteTarget;
    private static final LambdaForm.NamedFunction NF_checkCustomized;
    
    Invokers(MethodType targetType) {
        this.invokers = new MethodHandle[3];
        this.targetType = targetType;
    }
    
    MethodHandle exactInvoker() {
        MethodHandle cachedInvoker = this.cachedInvoker(0);
        if (cachedInvoker != null) {
            return cachedInvoker;
        }
        return this.setCachedInvoker(0, this.makeExactOrGeneralInvoker(true));
    }
    
    MethodHandle genericInvoker() {
        MethodHandle cachedInvoker = this.cachedInvoker(1);
        if (cachedInvoker != null) {
            return cachedInvoker;
        }
        return this.setCachedInvoker(1, this.makeExactOrGeneralInvoker(false));
    }
    
    MethodHandle basicInvoker() {
        MethodHandle cachedInvoker = this.cachedInvoker(2);
        if (cachedInvoker != null) {
            return cachedInvoker;
        }
        MethodType basicType = this.targetType.basicType();
        if (basicType != this.targetType) {
            return this.setCachedInvoker(2, basicType.invokers().basicInvoker());
        }
        MethodHandle methodHandle = basicType.form().cachedMethodHandle(0);
        if (methodHandle == null) {
            DirectMethodHandle make = DirectMethodHandle.make(invokeBasicMethod(basicType));
            assert !(!this.checkInvoker(make));
            methodHandle = basicType.form().setCachedMethodHandle(0, make);
        }
        return this.setCachedInvoker(2, methodHandle);
    }
    
    private MethodHandle cachedInvoker(int n) {
        return this.invokers[n];
    }
    
    private synchronized MethodHandle setCachedInvoker(int n, MethodHandle methodHandle) {
        MethodHandle methodHandle2 = this.invokers[n];
        if (methodHandle2 != null) {
            return methodHandle2;
        }
        return this.invokers[n] = methodHandle;
    }
    
    private MethodHandle makeExactOrGeneralInvoker(boolean b) {
        MethodType targetType = this.targetType;
        MethodHandle withInternalMemberName = BoundMethodHandle.bindSingle(targetType.invokerType(), invokeHandleForm(targetType, false, b ? 11 : 13), targetType).withInternalMemberName(MemberName.makeMethodHandleInvoke(b ? "invokeExact" : "invoke", targetType), false);
        assert !(!this.checkInvoker(withInternalMemberName));
        this.maybeCompileToBytecode(withInternalMemberName);
        return withInternalMemberName;
    }
    
    private void maybeCompileToBytecode(MethodHandle methodHandle) {
        if (this.targetType == this.targetType.erase() && this.targetType.parameterCount() < 10) {
            methodHandle.form.compileToBytecode();
        }
    }
    
    static MemberName invokeBasicMethod(MethodType obj) {
        assert obj == obj.basicType();
        try {
            return MethodHandles.Lookup.IMPL_LOOKUP.resolveOrFail((byte)5, MethodHandle.class, "invokeBasic", obj);
        }
        catch (ReflectiveOperationException ex) {
            throw MethodHandleStatics.newInternalError("JVM cannot find invoker for " + obj, ex);
        }
    }
    
    private boolean checkInvoker(MethodHandle methodHandle) {
        assert !(!this.targetType.invokerType().equals(methodHandle.type())) : Arrays.asList(new Object[] { this.targetType, this.targetType.invokerType(), methodHandle });
        assert !(!methodHandle.internalMemberName().getMethodType().equals(this.targetType));
        assert !methodHandle.isVarargsCollector();
        return true;
    }
    
    MethodHandle spreadInvoker(int n) {
        int i = this.targetType.parameterCount() - n;
        MethodType targetType = this.targetType;
        Class<Object> impliedRestargType = impliedRestargType(targetType, n);
        if (targetType.parameterSlotCount() <= 253) {
            return this.genericInvoker().asSpreader(impliedRestargType, i);
        }
        return MethodHandles.filterArgument(MethodHandles.invoker(targetType.replaceParameterTypes(n, targetType.parameterCount(), new Class[] { impliedRestargType })), 0, MethodHandles.insertArguments(Invokers.Lazy.MH_asSpreader, 1, new Object[] { impliedRestargType, Integer.valueOf(i) }));
    }
    
    private static Class<?> impliedRestargType(MethodType methodType, int n) {
        if (methodType.isGeneric()) {
            return Object[].class;
        }
        int parameterCount = methodType.parameterCount();
        if (n >= parameterCount) {
            return Object[].class;
        }
        Class<Object> parameterType = methodType.parameterType(n);
        int i = n + 1;
        while (i < parameterCount) {
            if (parameterType != methodType.parameterType(i)) {
                throw MethodHandleStatics.newIllegalArgumentException("need homogeneous rest arguments", methodType);
            }
            ++i;
        }
        if (parameterType == Object.class) {
            return Object[].class;
        }
        return Array.newInstance(parameterType, 0).getClass();
    }
    
    @Override
    public String toString() {
        return "Invokers" + this.targetType;
    }
    
    static MemberName methodHandleInvokeLinkerMethod(String str, MethodType methodType, Object[] array) {
        int n = -1;
        switch (str.hashCode()) {
            case 941760871: {
                if (str.equals("invokeExact")) {
                    n = 0;
                    break;
                }
                break;
            }
            case -1183693704: {
                if (str.equals("invoke")) {
                    n = 1;
                    break;
                }
                break;
            }
        }
        int n2;
        switch (n) {
            case 0: {
                n2 = 10;
                break;
            }
            case 1: {
                n2 = 12;
                break;
            }
            default: {
                throw new InternalError("not invoker: " + str);
            }
        }
        LambdaForm lambdaForm;
        if (methodType.parameterSlotCount() <= 253) {
            lambdaForm = invokeHandleForm(methodType, false, n2);
            array[0] = methodType;
        }
        else {
            lambdaForm = invokeHandleForm(methodType, true, n2);
        }
        return lambdaForm.vmentry;
    }
    
    private static LambdaForm invokeHandleForm(MethodType basicType, boolean b, int i) {
        int n;
        if (!b) {
            basicType = basicType.basicType();
            n = true;
        }
        else {
            n = false;
        }
        int n2;
        int n3;
        String s;
        switch (i) {
            case 10: {
                n2 = true;
                n3 = false;
                s = "invokeExact_MT";
                break;
            }
            case 11: {
                n2 = false;
                n3 = false;
                s = "exactInvoker";
                break;
            }
            case 12: {
                n2 = true;
                n3 = true;
                s = "invoke_MT";
                break;
            }
            case 13: {
                n2 = false;
                n3 = true;
                s = "invoker";
                break;
            }
            default: {
                throw new InternalError();
            }
        }
        if (!(n == 0)) {
            LambdaForm cachedLambdaForm = basicType.form().cachedLambdaForm(i);
            if (cachedLambdaForm != null) {
                return cachedLambdaForm;
            }
        }
        int from = 0 + (n2 ? 0 : 1);
        int n4 = from + 1 + basicType.parameterCount();
        int n5 = n4 + ((n2 && !b) ? 1 : 0);
        int n6 = b ? -1 : n4++;
        int n7 = n4++;
        int n8 = (MethodHandleStatics.CUSTOMIZE_THRESHOLD >= 0) ? n4++ : -1;
        int n9 = n4++;
        MethodType methodType = basicType.invokerType();
        if (!(n2 == 0)) {
            if (!b) {
                methodType = methodType.appendParameterTypes(new Class[] { MemberName.class });
            }
        }
        else {
            methodType = methodType.invokerType();
        }
        LambdaForm.Name[] arguments = LambdaForm.arguments(n4 - n5, methodType);
        assert arguments.length == n4 : Arrays.asList(new Serializable[] { basicType, Boolean.valueOf(b), Integer.valueOf(i), Integer.valueOf(n4), Integer.valueOf(arguments.length) });
        if (n6 >= n5) {
            assert arguments[n6] == null;
            BoundMethodHandle.SpeciesData speciesData_L = BoundMethodHandle.speciesData_L();
            arguments[0] = arguments[0].withConstraint(speciesData_L);
            arguments[n6] = new LambdaForm.Name(speciesData_L.getterFunction(0), new Object[] { arguments[0] });
        }
        MethodType basicType2 = basicType.basicType();
        Object[] copyOfRange = Arrays.copyOfRange(arguments, from, n4, Object[].class);
        Object o = b ? basicType : arguments[n6];
        if (n3 == 0) {
            arguments[n7] = new LambdaForm.Name(Invokers.NF_checkExactType, new Object[] { arguments[from], o });
        }
        else {
            copyOfRange[0] = (arguments[n7] = new LambdaForm.Name(Invokers.NF_checkGenericType, new Object[] { arguments[from], o }));
        }
        if (n8 != -1) {
            arguments[n8] = new LambdaForm.Name(Invokers.NF_checkCustomized, new Object[] { copyOfRange[0] });
        }
        arguments[n9] = new LambdaForm.Name(basicType2, copyOfRange);
        LambdaForm setCachedLambdaForm = new LambdaForm(s, n5, arguments);
        if (!(n2 == 0)) {
            setCachedLambdaForm.compileToBytecode();
        }
        if (!(n == 0)) {
            setCachedLambdaForm = basicType.form().setCachedLambdaForm(i, setCachedLambdaForm);
            return setCachedLambdaForm;
        }
        return setCachedLambdaForm;
    }
    
    static WrongMethodTypeException newWrongMethodTypeException(MethodType obj, MethodType obj2) {
        return new WrongMethodTypeException("expected " + obj2 + " but found " + obj);
    }
    
    @ForceInline
    static void checkExactType(Object o, Object o2) {
        MethodHandle methodHandle = (MethodHandle)o;
        MethodType methodType = (MethodType)o2;
        MethodType type = methodHandle.type();
        if (type != methodType) {
            throw newWrongMethodTypeException(methodType, type);
        }
        return;
    }
    
    @ForceInline
    static Object checkGenericType(Object o, Object o2) {
        return ((MethodHandle)o).asType((MethodType)o2);
    }
    
    static MemberName linkToCallSiteMethod(MethodType methodType) {
        return callSiteForm(methodType, false).vmentry;
    }
    
    static MemberName linkToTargetMethod(MethodType methodType) {
        return callSiteForm(methodType, true).vmentry;
    }
    
    private static LambdaForm callSiteForm(MethodType basicType, boolean b) {
        basicType = basicType.basicType();
        int n = b ? 15 : 14;
        LambdaForm cachedLambdaForm = basicType.form().cachedLambdaForm(n);
        if (cachedLambdaForm != null) {
            return cachedLambdaForm;
        }
        int n2 = 0 + basicType.parameterCount();
        int n3 = n2 + 1;
        int n4 = n2++;
        n4 = (b ? -1 : n4);
        int n5 = b ? n4 : n2++;
        int n6 = n2++;
        LambdaForm.Name[] arguments = LambdaForm.arguments(n2 - n3, basicType.appendParameterTypes(new Class[] { b ? MethodHandle.class : CallSite.class }));
        assert arguments.length == n2;
        assert arguments[n4] != null;
        if (!b) {
            arguments[n5] = new LambdaForm.Name(Invokers.NF_getCallSiteTarget, new Object[] { arguments[n4] });
        }
        Object[] copyOfRange = Arrays.copyOfRange(arguments, 0, n2 + 1, Object[].class);
        System.arraycopy(copyOfRange, 0, copyOfRange, 1, copyOfRange.length - 1);
        copyOfRange[0] = arguments[n5];
        arguments[n6] = new LambdaForm.Name(basicType, copyOfRange);
        LambdaForm o = new LambdaForm(b ? "linkToTargetMethod" : "linkToCallSite", n3, arguments);
        o.compileToBytecode();
        return basicType.form().setCachedLambdaForm(n, o);
    }
    
    @ForceInline
    static Object getCallSiteTarget(Object o) {
        return ((CallSite)o).getTarget();
    }
    
    @ForceInline
    static void checkCustomized(Object o) {
        MethodHandle methodHandle = (MethodHandle)o;
        if (methodHandle.form.customized == null) {
            maybeCustomize(methodHandle);
        }
    }
    
    @DontInline
    static void maybeCustomize(MethodHandle methodHandle) {
        int customizationCount = methodHandle.customizationCount;
        if (customizationCount >= MethodHandleStatics.CUSTOMIZE_THRESHOLD) {
            methodHandle.customize();
        }
        else {
            methodHandle.customizationCount = (byte)(customizationCount + 1);
        }
    }
    
    static {
        try {
            LambdaForm.NamedFunction[] array = new LambdaForm.NamedFunction[] { NF_checkExactType = new LambdaForm.NamedFunction(Invokers.class.getDeclaredMethod("checkExactType", new Class[] { Object.class, Object.class })), NF_checkGenericType = new LambdaForm.NamedFunction(Invokers.class.getDeclaredMethod("checkGenericType", new Class[] { Object.class, Object.class })), NF_getCallSiteTarget = new LambdaForm.NamedFunction(Invokers.class.getDeclaredMethod("getCallSiteTarget", new Class[] { Object.class })), NF_checkCustomized = new LambdaForm.NamedFunction(Invokers.class.getDeclaredMethod("checkCustomized", new Class[] { Object.class })) };
            int length = array.length;
            int i = 0;
            while (i < length) {
                LambdaForm.NamedFunction detailMessage = array[i];
                assert !(!InvokerBytecodeGenerator.isStaticallyInvocable(detailMessage.member)) : detailMessage;
                detailMessage.resolve();
                ++i;
            }
        }
        catch (ReflectiveOperationException ex) {
            throw MethodHandleStatics.newInternalError(ex);
        }
    }
    
    private static class Lazy
    {
        private static final MethodHandle MH_asSpreader;
        
        static {
            try {
                MH_asSpreader = MethodHandles.Lookup.IMPL_LOOKUP.findVirtual(MethodHandle.class, "asSpreader", MethodType.methodType(MethodHandle.class, Class.class, new Class[] { Integer.TYPE }));
            }
            catch (ReflectiveOperationException ex) {
                throw MethodHandleStatics.newInternalError(ex);
            }
        }
    }
}
